# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("wsdapi.dll")
prototypes = \
    {
        # 
        'WSDXMLGetNameFromBuiltinNamespace': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszNamespace", "pszName", "ppName"]),
        # 
        'WSDXMLCreateContext': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWSDXMLContext"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppContext"]),
        # 
        'WSDCreateUdpMessageParameters': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWSDUdpMessageParameters"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppTxParams"]),
        # 
        'WSDCreateUdpAddress': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWSDUdpAddress"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppAddress"]),
        # 
        'WSDCreateHttpMessageParameters': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWSDHttpMessageParameters"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppTxParams"]),
        # 
        'WSDCreateHttpAddress': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWSDHttpAddress"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppAddress"]),
        # 
        'WSDSetConfigurationOption': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["dwOption", "pVoid", "cbInBuffer"]),
        # 
        'WSDGetConfigurationOption': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["dwOption", "pVoid", "cbOutBuffer"]),
        # 
        'WSDAllocateLinkedMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["pParent", "cbSize"]),
        # 
        'WSDFreeLinkedMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pVoid"]),
        # 
        'WSDAttachLinkedMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pParent", "pChild"]),
        # 
        'WSDDetachLinkedMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pVoid"]),
        # 
        'WSDXMLBuildAnyForSingleElement': SimTypeFunction([SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pElementName", "pszText", "ppAny"]),
        # 
        'WSDXMLGetValueFromAny': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszNamespace", "pszName", "pAny", "ppszValue"]),
        # 
        'WSDXMLAddSibling': SimTypeFunction([SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pFirst", "pSecond"]),
        # 
        'WSDXMLAddChild': SimTypeFunction([SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pParent", "pChild"]),
        # 
        'WSDXMLCleanupElement': SimTypeFunction([SimTypePointer(SimStruct({"Node": SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "FirstAttribute": SimTypePointer(SimStruct({"Element": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_ATTRIBUTE"), offset=0), "Name": SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), "Value": SimTypeBottom(label="PWSTR")}, name="WSDXML_ATTRIBUTE", pack=False, align=None), offset=0), "FirstChild": SimTypePointer(SimStruct({"Type": SimTypeInt(signed=True, label="Int32"), "Parent": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_NODE"), offset=0)}, name="WSDXML_NODE", pack=False, align=None), offset=0), "PrefixMappings": SimTypePointer(SimStruct({"Refs": SimTypeInt(signed=False, label="UInt32"), "Next": SimTypePointer(SimTypeBottom(label="WSDXML_PREFIX_MAPPING"), offset=0), "Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "Prefix": SimTypeBottom(label="PWSTR")}, name="WSDXML_PREFIX_MAPPING", pack=False, align=None), offset=0)}, name="WSDXML_ELEMENT", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pAny"]),
        # 
        'WSDGenerateFault': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypePointer(SimStruct({"Code": SimTypePointer(SimStruct({"Value": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "Subcode": SimTypePointer(SimStruct({"Value": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "Subcode": SimTypePointer(SimTypeBottom(label="WSD_SOAP_FAULT_SUBCODE"), offset=0)}, name="WSD_SOAP_FAULT_SUBCODE", pack=False, align=None), offset=0)}, name="WSD_SOAP_FAULT_CODE", pack=False, align=None), offset=0), "Reason": SimTypePointer(SimStruct({"Text": SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="WSD_LOCALIZED_STRING_LIST"), offset=0), "Element": SimTypePointer(SimTypeBottom(label="WSD_LOCALIZED_STRING"), offset=0)}, name="WSD_LOCALIZED_STRING_LIST", pack=False, align=None), offset=0)}, name="WSD_SOAP_FAULT_REASON", pack=False, align=None), offset=0), "Node": SimTypeBottom(label="PWSTR"), "Role": SimTypeBottom(label="PWSTR"), "Detail": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0)}, name="WSD_SOAP_FAULT", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszCode", "pszSubCode", "pszReason", "pszDetail", "pContext", "ppFault"]),
        # 
        'WSDGenerateFaultEx': SimTypeFunction([SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Space": SimTypePointer(SimStruct({"Uri": SimTypeBottom(label="PWSTR"), "PreferredPrefix": SimTypeBottom(label="PWSTR"), "Names": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "NamesCount": SimTypeShort(signed=False, label="UInt16"), "Encoding": SimTypeShort(signed=False, label="UInt16")}, name="WSDXML_NAMESPACE", pack=False, align=None), offset=0), "LocalName": SimTypeBottom(label="PWSTR")}, name="WSDXML_NAME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="WSD_LOCALIZED_STRING_LIST"), offset=0), "Element": SimTypePointer(SimTypeBottom(label="WSD_LOCALIZED_STRING"), offset=0)}, name="WSD_LOCALIZED_STRING_LIST", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"Code": SimTypePointer(SimStruct({"Value": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "Subcode": SimTypePointer(SimStruct({"Value": SimTypePointer(SimTypeBottom(label="WSDXML_NAME"), offset=0), "Subcode": SimTypePointer(SimTypeBottom(label="WSD_SOAP_FAULT_SUBCODE"), offset=0)}, name="WSD_SOAP_FAULT_SUBCODE", pack=False, align=None), offset=0)}, name="WSD_SOAP_FAULT_CODE", pack=False, align=None), offset=0), "Reason": SimTypePointer(SimStruct({"Text": SimTypePointer(SimStruct({"Next": SimTypePointer(SimTypeBottom(label="WSD_LOCALIZED_STRING_LIST"), offset=0), "Element": SimTypePointer(SimTypeBottom(label="WSD_LOCALIZED_STRING"), offset=0)}, name="WSD_LOCALIZED_STRING_LIST", pack=False, align=None), offset=0)}, name="WSD_SOAP_FAULT_REASON", pack=False, align=None), offset=0), "Node": SimTypeBottom(label="PWSTR"), "Role": SimTypeBottom(label="PWSTR"), "Detail": SimTypePointer(SimTypeBottom(label="WSDXML_ELEMENT"), offset=0)}, name="WSD_SOAP_FAULT", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pCode", "pSubCode", "pReasons", "pszDetail", "ppFault"]),
        # 
        'WSDUriEncode': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["source", "cchSource", "destOut", "cchDestOut"]),
        # 
        'WSDUriDecode': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["source", "cchSource", "destOut", "cchDestOut"]),
        # 
        'WSDCreateDeviceHost': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypeBottom(label="IWSDDeviceHost"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszLocalId", "pContext", "ppDeviceHost"]),
        # 
        'WSDCreateDeviceHostAdvanced': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypeBottom(label="IWSDAddress"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWSDDeviceHost"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszLocalId", "pContext", "ppHostAddresses", "dwHostAddressCount", "ppDeviceHost"]),
        # 
        'WSDCreateDeviceHost2': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimStruct({"configParamType": SimTypeInt(signed=False, label="WSD_CONFIG_PARAM_TYPE"), "pConfigData": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwConfigDataSize": SimTypeInt(signed=False, label="UInt32")}, name="WSD_CONFIG_PARAM", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWSDDeviceHost"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszLocalId", "pContext", "pConfigParams", "dwConfigParamCount", "ppDeviceHost"]),
        # 
        'WSDCreateDeviceProxy': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypeBottom(label="IWSDDeviceProxy"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszDeviceId", "pszLocalId", "pContext", "ppDeviceProxy"]),
        # 
        'WSDCreateDeviceProxyAdvanced': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDAddress"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypeBottom(label="IWSDDeviceProxy"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszDeviceId", "pDeviceAddress", "pszLocalId", "pContext", "ppDeviceProxy"]),
        # 
        'WSDCreateDeviceProxy2': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimStruct({"configParamType": SimTypeInt(signed=False, label="WSD_CONFIG_PARAM_TYPE"), "pConfigData": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwConfigDataSize": SimTypeInt(signed=False, label="UInt32")}, name="WSD_CONFIG_PARAM", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWSDDeviceProxy"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszDeviceId", "pszLocalId", "pContext", "pConfigParams", "dwConfigParamCount", "ppDeviceProxy"]),
        # 
        'WSDCreateDiscoveryProvider': SimTypeFunction([SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypeBottom(label="IWSDiscoveryProvider"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pContext", "ppProvider"]),
        # 
        'WSDCreateDiscoveryProvider2': SimTypeFunction([SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimStruct({"configParamType": SimTypeInt(signed=False, label="WSD_CONFIG_PARAM_TYPE"), "pConfigData": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwConfigDataSize": SimTypeInt(signed=False, label="UInt32")}, name="WSD_CONFIG_PARAM", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWSDiscoveryProvider"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pContext", "pConfigParams", "dwConfigParamCount", "ppProvider"]),
        # 
        'WSDCreateDiscoveryPublisher': SimTypeFunction([SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimTypeBottom(label="IWSDiscoveryPublisher"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pContext", "ppPublisher"]),
        # 
        'WSDCreateDiscoveryPublisher2': SimTypeFunction([SimTypeBottom(label="IWSDXMLContext"), SimTypePointer(SimStruct({"configParamType": SimTypeInt(signed=False, label="WSD_CONFIG_PARAM_TYPE"), "pConfigData": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "dwConfigDataSize": SimTypeInt(signed=False, label="UInt32")}, name="WSD_CONFIG_PARAM", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IWSDiscoveryPublisher"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pContext", "pConfigParams", "dwConfigParamCount", "ppPublisher"]),
        # 
        'WSDCreateOutboundAttachment': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IWSDOutboundAttachment"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppAttachment"]),
    }

lib.set_prototypes(prototypes)



from typing import Dict, Tuple, Union, Optional

import claripy
import ailment

from ...calling_conventions import SimFunctionArgument, SimRegArg
from ...engines.light import SpOffset
from .heap_address import HeapAddress
from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues


class Atom:
    """
    This class represents a data storage location manipulated by IR instructions.

    It could either be a Tmp (temporary variable), a Register, a MemoryLocation.
    """
    def __repr__(self):
        raise NotImplementedError()

    @property
    def size(self) -> int:
        raise NotImplementedError()

    @staticmethod
    def from_argument(argument: SimFunctionArgument, registers: Dict[str,Tuple[int,int]]):
        """
        Instanciate an `Atom` from a given argument.

        :param argument: The argument to create a new atom from.
        :param registers: A mapping representing the registers of a given architecture.
        """
        if isinstance(argument, SimRegArg):
            return Register(registers[argument.reg_name][0], argument.size)
        else:
            raise TypeError("Argument type %s is not yet supported." % type(argument))


class GuardUse(Atom):
    """
    Implements a guard use.
    """
    def __init__(self, target):
        self.target = target

    def __repr__(self):
        return '<Guard %#x>' % self.target

    @property
    def size(self) -> int:
        raise NotImplementedError()


class FunctionCall(Atom):
    def __init__(self, target, callsite):
        self.target = target
        self.callsite = callsite

    @property
    def single_target(self):
        if type(self.target) is MultiValues and len(self.target.values) == 1 and 0 in self.target.values and \
                len(self.target.values[0]) == 1 and next(iter(self.target.values[0])).op == 'BVV':
            return next(iter(self.target.values[0])).args[0]
        return None

    def __repr__(self):
        target = self.single_target
        target_txt = hex(target) if target is not None else '(indirect)'
        return '<Call %s>' % target_txt

    def __eq__(self, other):
        return type(other) is FunctionCall and self.callsite == other.callsite

    def __hash__(self):
        return hash(self.callsite)

    @property
    def size(self):
        raise NotImplementedError


class ConstantSrc(Atom):
    def __init__(self, const):
        self.const = const

    def __repr__(self):
        return repr(self.const)

    def __eq__(self, other):
        return type(other) is ConstantSrc and self.const == other.const

    def __hash__(self):
        return hash(self.const)

    @property
    def size(self):
        return self.const.size


class Tmp(Atom):
    """
    Represents a variable used by the IR to store intermediate values.
    """
    __slots__ = ('tmp_idx', '_size', )

    def __init__(self, tmp_idx: int, size: int):
        super().__init__()
        self.tmp_idx = tmp_idx
        self._size = size

    def __repr__(self):
        return "<Tmp %d>" % self.tmp_idx

    def __eq__(self, other):
        return type(other) is Tmp and self.tmp_idx == other.tmp_idx

    def __hash__(self):
        return hash(('tmp', self.tmp_idx))

    @property
    def size(self) -> int:
        return self._size


class Register(Atom):
    """
    Represents a given CPU register.

    As an IR abstracts the CPU design to target different architectures, registers are represented as a separated memory
    space.
    Thus a register is defined by its offset from the base of this memory and its size.

    :ivar int reg_offset:    The offset from the base to define its place in the memory bloc.
    :ivar int size:          The size, in number of bytes.
    """
    __slots__ = ('reg_offset', '_size', )

    def __init__(self, reg_offset: int, size: int):
        super().__init__()

        self.reg_offset = reg_offset
        self._size = size

    def __repr__(self):
        return "<Reg %d<%d>>" % (self.reg_offset, self.size)

    def __eq__(self, other):
        return type(other) is Register and \
               self.reg_offset == other.reg_offset and \
               self.size == other.size

    def __hash__(self):
        return hash(('reg', self.reg_offset, self.size))

    @property
    def bits(self) -> int:
        return self._size * 8

    @property
    def size(self) -> int:
        return self._size


class MemoryLocation(Atom):
    """
    Represents a memory slice.

    It is characterized by its address and its size.
    """

    __slots__ = ('addr', '_size', 'endness', )

    def __init__(self, addr: Union[SpOffset,HeapAddress,int], size: int, endness: Optional[str]=None):
        """
        :param int addr: The address of the beginning memory location slice.
        :param int size: The size of the represented memory location, in bytes.
        """
        super().__init__()

        self.addr: Union[SpOffset,int,claripy.ast.BV] = addr
        self._size: int = size
        self.endness = endness

    def __repr__(self):
        address_format = hex(self.addr) if type(self.addr) is int else self.addr
        stack_format = ' (stack)' if self.is_on_stack else ''
        size = "%d" % self.size if isinstance(self.size, int) else self.size

        return "<Mem %s<%s>%s>" % (address_format, size, stack_format)

    @property
    def is_on_stack(self) -> bool:
        """
        True if this memory location is located on the stack.
        """
        return isinstance(self.addr, SpOffset)

    @property
    def bits(self) -> int:
        return self.size * 8

    @property
    def size(self) -> int:
        return self._size

    @property
    def symbolic(self) -> bool:
        if isinstance(self.addr, int):
            return False
        elif isinstance(self.addr, SpOffset):
            return not type(self.addr.offset) is int
        return True

    def __eq__(self, other):
        # pylint:disable=isinstance-second-argument-not-valid-type
        return type(other) is MemoryLocation and \
               (
                    self.addr is other.addr if isinstance(self.addr, (claripy.ast.BV)) else self.addr == other.addr
               ) and \
               self.size == other.size and \
               self.endness == other.endness

    def __hash__(self):
        return hash(('mem', self.addr, self.size, self.endness))



class Undefined:
    """
    A TOP-like value indicating an unknown data source. Should live next to raw integers in DataSets.
    """
    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __lshift__(self, other):
        return self

    def __rlshift__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __rrshift__(self, other):
        return self

    def __and__(self, other):
        return self

    def __rand__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __rxor__(self, other):
        return self

    def __or__(self, other):
        return self

    def __ror__(self, other):
        return self

    def __neg__(self):
        return self

    def __eq__(self, other):
        return type(other) is Undefined

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash('undefined')

    def __str__(self):
        return '<Undefined>'

    def __repr__(self):
        return "<Undefined>"

UNDEFINED = Undefined()



from typing import Optional

from ailment import Block, Stmt

from ..decompiler.ailblock_walker import AILBlockWalker


class CallExprFinder(AILBlockWalker):
    """
    Walks an AIL expression to find if it contains a call expression anywhere.
    """
    def __init__(self):
        super().__init__()
        self.has_call = False

    # pylint:disable=unused-argument
    def _handle_CallExpr(self, expr_idx: int, expr: Stmt.Call, stmt_idx: int, stmt: Stmt.Statement,
                         block: Optional[Block]):
        self.has_call = True



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("faultrep.dll")
prototypes = \
    {
        # 
        'ReportFault': SimTypeFunction([SimTypePointer(SimStruct({"ExceptionRecord": SimTypePointer(SimStruct({"ExceptionCode": SimTypeInt(signed=True, label="Int32"), "ExceptionFlags": SimTypeInt(signed=False, label="UInt32"), "ExceptionRecord": SimTypePointer(SimTypeBottom(label="EXCEPTION_RECORD"), offset=0), "ExceptionAddress": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "NumberParameters": SimTypeInt(signed=False, label="UInt32"), "ExceptionInformation": SimTypeFixedSizeArray(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), 15)}, name="EXCEPTION_RECORD", pack=False, align=None), offset=0), "ContextRecord": SimTypePointer(SimStruct({"P1Home": SimTypeLongLong(signed=False, label="UInt64"), "P2Home": SimTypeLongLong(signed=False, label="UInt64"), "P3Home": SimTypeLongLong(signed=False, label="UInt64"), "P4Home": SimTypeLongLong(signed=False, label="UInt64"), "P5Home": SimTypeLongLong(signed=False, label="UInt64"), "P6Home": SimTypeLongLong(signed=False, label="UInt64"), "ContextFlags": SimTypeInt(signed=False, label="UInt32"), "MxCsr": SimTypeInt(signed=False, label="UInt32"), "SegCs": SimTypeShort(signed=False, label="UInt16"), "SegDs": SimTypeShort(signed=False, label="UInt16"), "SegEs": SimTypeShort(signed=False, label="UInt16"), "SegFs": SimTypeShort(signed=False, label="UInt16"), "SegGs": SimTypeShort(signed=False, label="UInt16"), "SegSs": SimTypeShort(signed=False, label="UInt16"), "EFlags": SimTypeInt(signed=False, label="UInt32"), "Dr0": SimTypeLongLong(signed=False, label="UInt64"), "Dr1": SimTypeLongLong(signed=False, label="UInt64"), "Dr2": SimTypeLongLong(signed=False, label="UInt64"), "Dr3": SimTypeLongLong(signed=False, label="UInt64"), "Dr6": SimTypeLongLong(signed=False, label="UInt64"), "Dr7": SimTypeLongLong(signed=False, label="UInt64"), "Rax": SimTypeLongLong(signed=False, label="UInt64"), "Rcx": SimTypeLongLong(signed=False, label="UInt64"), "Rdx": SimTypeLongLong(signed=False, label="UInt64"), "Rbx": SimTypeLongLong(signed=False, label="UInt64"), "Rsp": SimTypeLongLong(signed=False, label="UInt64"), "Rbp": SimTypeLongLong(signed=False, label="UInt64"), "Rsi": SimTypeLongLong(signed=False, label="UInt64"), "Rdi": SimTypeLongLong(signed=False, label="UInt64"), "R8": SimTypeLongLong(signed=False, label="UInt64"), "R9": SimTypeLongLong(signed=False, label="UInt64"), "R10": SimTypeLongLong(signed=False, label="UInt64"), "R11": SimTypeLongLong(signed=False, label="UInt64"), "R12": SimTypeLongLong(signed=False, label="UInt64"), "R13": SimTypeLongLong(signed=False, label="UInt64"), "R14": SimTypeLongLong(signed=False, label="UInt64"), "R15": SimTypeLongLong(signed=False, label="UInt64"), "Rip": SimTypeLongLong(signed=False, label="UInt64"), "Anonymous": SimUnion({"FltSave": SimTypeBottom(label="XSAVE_FORMAT"), "Anonymous": SimStruct({"Header": SimTypeFixedSizeArray(SimTypeBottom(label="M128A"), 2), "Legacy": SimTypeFixedSizeArray(SimTypeBottom(label="M128A"), 8), "Xmm0": SimTypeBottom(label="M128A"), "Xmm1": SimTypeBottom(label="M128A"), "Xmm2": SimTypeBottom(label="M128A"), "Xmm3": SimTypeBottom(label="M128A"), "Xmm4": SimTypeBottom(label="M128A"), "Xmm5": SimTypeBottom(label="M128A"), "Xmm6": SimTypeBottom(label="M128A"), "Xmm7": SimTypeBottom(label="M128A"), "Xmm8": SimTypeBottom(label="M128A"), "Xmm9": SimTypeBottom(label="M128A"), "Xmm10": SimTypeBottom(label="M128A"), "Xmm11": SimTypeBottom(label="M128A"), "Xmm12": SimTypeBottom(label="M128A"), "Xmm13": SimTypeBottom(label="M128A"), "Xmm14": SimTypeBottom(label="M128A"), "Xmm15": SimTypeBottom(label="M128A")}, name="_Anonymous_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "VectorRegister": SimTypeFixedSizeArray(SimTypeBottom(label="M128A"), 26), "VectorControl": SimTypeLongLong(signed=False, label="UInt64"), "DebugControl": SimTypeLongLong(signed=False, label="UInt64"), "LastBranchToRip": SimTypeLongLong(signed=False, label="UInt64"), "LastBranchFromRip": SimTypeLongLong(signed=False, label="UInt64"), "LastExceptionToRip": SimTypeLongLong(signed=False, label="UInt64"), "LastExceptionFromRip": SimTypeLongLong(signed=False, label="UInt64")}, name="CONTEXT", pack=False, align=None), offset=0)}, name="EXCEPTION_POINTERS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="EFaultRepRetVal"), arg_names=["pep", "dwOpt"]),
        # 
        'AddERExcludedApplicationA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["szApplication"]),
        # 
        'AddERExcludedApplicationW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wszApplication"]),
        # 
        'WerReportHang': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwndHungApp", "pwzHungApplicationName"]),
    }

lib.set_prototypes(prototypes)



import collections.abc
import contextlib
import threading
import tempfile
import weakref
import logging
import pickle
import shelve
import uuid
import os
import io

import claripy

l = logging.getLogger("angr.vault")

class VaultPickler(pickle.Pickler):
    def __init__(self, vault, file, *args, assigned_objects=(), **kwargs):
        """
        A persistence-aware pickler.
        It will check for persistence of any objects except for those with IDs in 'assigned_objects'.
        """

        super().__init__(file, *args, **kwargs)
        self.vault = vault
        self.assigned_objects = assigned_objects

    def persistent_id(self, obj):
        if any(obj is o for o in self.assigned_objects):
            return None

        pid = self.vault._get_persistent_id(obj)
        if pid is None:
            return None

        # l.debug("Persistent store: %s %s", obj, pid)
        return self.vault._store(obj, pid)

class VaultUnpickler(pickle.Unpickler):
    def __init__(self, vault, file, *args, **kwargs):
        super().__init__(file, *args, **kwargs)
        self.vault = vault

    def persistent_load(self, pid):
        return self.vault._load(pid)

class Vault(collections.abc.MutableMapping):
    """
    The vault is a serializer for angr.
    """

    #
    # These MUST be overriden.
    #

    def _read_context(self, i):
        """
        Should be a context that yields a pickle-read()able file object for the given id i.
        """
        raise NotImplementedError()

    def _write_context(self, i):
        """
        Should be a context that yields a pickle-write()able file object for the given id i.
        """
        raise NotImplementedError()

    def keys(self):
        """
        Should return the IDs stored by the vault.
        """
        raise NotImplementedError()

    #
    # Persistance managers
    #

    def __init__(self):
        self._object_cache = weakref.WeakValueDictionary()
        self._uuid_cache = weakref.WeakKeyDictionary()
        self.stored = set()
        self.storing = set()
        self.hash_dedup = {
            claripy.ast.Base, claripy.ast.BV, claripy.ast.FP, claripy.ast.Bool, claripy.ast.Int, claripy.ast.Bits,
        }
        self.module_dedup = set() # {'claripy', 'angr', 'archinfo', 'pyvex' } # cle causes recursion
        self.uuid_dedup = { SimState, Project }
        self.unsafe_key_baseclasses = {
            claripy.ast.Base, SimType
        }

    def _get_persistent_id(self, o):
        """
        Determines a persistent ID for an object.
        Does NOT do stores.
        """
        if type(o) in self.hash_dedup:
            oid = o.__class__.__name__ + "-" + str(hash(o))
            self._object_cache[oid] = o
            return oid

        if any(isinstance(o,c) for c in self.unsafe_key_baseclasses):
            return None

        try:
            return self._uuid_cache[o]
        except KeyError:
            pass
        except TypeError:
            return None

        #if type(o) in self.uuid_dedup:
        #    return self._get_id(o)
        if o.__class__.__module__.split('.')[0] in self.module_dedup or o.__class__ in self.uuid_dedup:
            oid = o.__class__.__name__.split(".")[-1] + '-' + str(uuid.uuid4())
            self._object_cache[oid] = o
            self._uuid_cache[o] = oid
            return oid

        return None

    #
    # Other stuff
    #

    def is_stored(self, i):
        """
        Checks if the provided id is already in the vault.
        """
        if i in self.stored:
            return True

        try:
            with self._read_context(i):
                return True
        except (AngrVaultError, EOFError):
            return False

    def load(self, oid):
        return self._load(oid)

    def _load(self, oid):
        """
        Retrieves one object from the pickler with the provided id.

        :param oid: an ID to use
        """
        # l.debug("LOAD: %s", oid)
        try:
            # l.debug("... trying cached")
            return self._object_cache[oid]
        except KeyError:
            # l.debug("... cached failed")
            with self._read_context(oid) as u:
                # add newly loaded object into the object cache
                o = VaultUnpickler(self, u).load()
                self._object_cache[oid] = o
                return o

    def store(self, o):

        actual_id = self._get_persistent_id(o) or "TMP-"+str(uuid.uuid4())

        return self._store(o, actual_id)

    def _store(self, o, oid): #pylint:disable=redefined-builtin
        """
        Stores an object and returns its ID.

        :param o: the object
        :param oid: an ID to use
        """

        actual_id = oid

        # l.debug("STORE: %s %s", o, actual_id)

        # this handles recursive objects
        if actual_id in self.storing:
            return actual_id

        if self.is_stored(actual_id):
            # l.debug("... already stored")
            return actual_id

        with self._write_context(actual_id) as output:
            self.storing.add(actual_id)
            VaultPickler(self, output, assigned_objects=(o,)).dump(o)
            self.stored.add(actual_id)

        return actual_id

    def dumps(self, o):
        """
        Returns a serialized string representing the object, post-deduplication.

        :param o: the object
        """
        f = io.BytesIO()
        VaultPickler(self, f).dump(o)
        f.seek(0)
        return f.read()

    def loads(self, s):
        """
        Deserializes a string representation of the object.

        :param s: the string
        """
        f = io.BytesIO(s)
        return VaultUnpickler(self, f).load()

    def _clear_cache(self):

        self._object_cache.clear()
        self._uuid_cache.clear()
        self.stored.clear()
        self.storing.clear()

    @staticmethod
    def close():
        pass

    #
    # For MutableMapping.
    #

    def __setitem__(self, k, v):
        self._store(v, k)

    def __getitem__(self, k):
        return self.load(k)

    def __delitem__(self, k):
        raise AngrVaultError("We currently don't support deletion from the vault.")

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

class VaultDict(Vault):
    """
    A Vault that uses a dictionary for storage.
    """
    def __init__(self, d=None):
        super().__init__()
        self._dict = { } if d is None else d

    @contextlib.contextmanager
    def _write_context(self, i):
        f = io.BytesIO()
        yield f
        f.seek(0)
        self._dict[i] = f.read()

    @contextlib.contextmanager
    def _read_context(self, i):
        try:
            f = io.BytesIO(self._dict[i])
            yield f
        except KeyError as e:
            raise AngrVaultError from e

    def is_stored(self, i):
        return i in self._dict

    def keys(self):
        return self._dict.keys()

class VaultDir(Vault):
    """
    A Vault that uses a directory for storage.
    """
    def __init__(self, d=None):
        super().__init__()
        self._dir = tempfile.mkdtemp() if d is None else d
        with contextlib.suppress(FileExistsError):
            os.makedirs(self._dir)

    @contextlib.contextmanager
    def _write_context(self, i):
        with open(os.path.join(self._dir, i), "wb") as o:
            yield o

    @contextlib.contextmanager
    def _read_context(self, i):
        try:
            with open(os.path.join(self._dir, i), "rb") as o:
                yield o
        except FileNotFoundError as e:
            raise AngrVaultError from e

    def keys(self):
        return os.listdir(self._dir)


class VaultShelf(VaultDict):
    """
    A Vault that uses a shelve.Shelf for storage.
    """
    def __init__(self, path=None):
        self._path = tempfile.mktemp() if path is None else path
        s = shelve.open(self._path, protocol=-1)
        super().__init__(s)

    def close(self):
        self._dict.close()


class VaultDirShelf(VaultDict):
    """
    A Vault that uses a directory for storage, where every object is stored into a single shelve.Shelf instance.
    VaultDir creates a file for each object. VaultDirShelf creates only one file for a stored object and everything
    else it references.
    """

    def __init__(self, d=None):
        super().__init__()
        self._d = tempfile.mkdtemp() if d is None else d
        self._dict = None  # will be initialized at each call to store() or load()
        self._lock = threading.Lock()  # protecting access to self._dict

    @contextlib.contextmanager
    def _locked_shelve(self, shelve_path):
        with self._lock:
            self._dict = shelve.open(shelve_path, protocol=-1)
            yield
            self._dict.close()
            self._dict = None
            self._clear_cache()

    def store(self, o):
        oid = self._get_persistent_id(o) or "TMP-"+str(uuid.uuid4())
        shelve_path = os.path.join(self._d, oid)
        with self._locked_shelve(shelve_path):
            super()._store(o, oid)
        return oid

    def load(self, oid):

        shelve_path = os.path.join(self._d, oid)
        with self._locked_shelve(shelve_path):
            o = super().load(oid)

        return o

    def keys(self):
        s = set()
        for n in os.listdir(self._d):
            if "." not in n:
                s.add(n)
            else:
                s.add(n[:n.rfind(".")])  # remove the suffix
        return s


from .errors import AngrVaultError
from .project import Project
from .sim_type import SimType
from .sim_state import SimState



from itertools import chain
from typing import Iterable, Optional
import logging

import archinfo
import claripy
import ailment
import pyvex
from claripy import FSORT_DOUBLE, FSORT_FLOAT

from ...engines.light import SimEngineLight, SimEngineLightAILMixin, SpOffset
from ...errors import SimEngineError, SimMemoryMissingError
from ...calling_conventions import DEFAULT_CC, SimRegArg, SimStackArg
from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from ...knowledge_plugins.key_definitions.atoms import Register, Tmp, MemoryLocation
from ...knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
from ...knowledge_plugins.key_definitions.live_definitions import Definition
from .subject import SubjectType
from .external_codeloc import ExternalCodeLocation
from .rd_state import ReachingDefinitionsState

l = logging.getLogger(name=__name__)


class SimEngineRDAIL(
    SimEngineLightAILMixin,
    SimEngineLight,
):  # pylint:disable=abstract-method

    arch: archinfo.Arch
    state: ReachingDefinitionsState

    def __init__(self, project, call_stack, maximum_local_call_depth, function_handler=None):
        super().__init__()
        self.project = project
        self._call_stack = call_stack
        self._maximum_local_call_depth = maximum_local_call_depth
        self._function_handler = function_handler
        self._visited_blocks = None
        self._dep_graph = None

    def process(self, state, *args, **kwargs):
        self._dep_graph = kwargs.pop('dep_graph', None)
        self._visited_blocks = kwargs.pop('visited_blocks', None)

        # we are using a completely different state. Therefore, we directly call our _process() method before
        # SimEngine becomes flexible enough.
        try:
            self._process(
                state,
                None,
                block=kwargs.pop('block', None),
            )
        except SimEngineError as e:
            if kwargs.pop('fail_fast', False) is True:
                raise e
        return self.state, self._visited_blocks, self._dep_graph

    def sp_offset(self, offset: int):
        return self.state.stack_address(offset)

    #
    # Private methods
    #

    @staticmethod
    def _external_codeloc():
        return ExternalCodeLocation()

    #
    # AIL statement handlers
    #

    def _handle_Stmt(self, stmt):

        if self.state.analysis:
            self.state.analysis.insn_observe(self.ins_addr, stmt, self.block, self.state, OP_BEFORE)

        super()._handle_Stmt(stmt)

        if self.state.analysis:
            self.state.analysis.insn_observe(self.ins_addr, stmt, self.block, self.state, OP_AFTER)

    def _ail_handle_Assignment(self, stmt):
        """

        :param ailment.Assignment stmt:
        :return:
        """

        src = self._expr(stmt.src)
        dst = stmt.dst

        if src is None:
            src = MultiValues(offset_to_values={0: {self.state.top(dst.bits)}})

        if isinstance(dst, ailment.Tmp):
            self.state.kill_and_add_definition(Tmp(dst.tmp_idx, dst.size), self._codeloc(), src)
            self.tmps[dst.tmp_idx] = src

        elif isinstance(dst, ailment.Register):
            reg = Register(dst.reg_offset, dst.size)
            self.state.kill_and_add_definition(reg, self._codeloc(), src)

            if dst.reg_offset == self.arch.sp_offset:
                # TODO: Special logic that frees all definitions above the current stack pointer
                pass

        else:
            l.warning('Unsupported type of Assignment dst %s.', type(dst).__name__)

    def _ail_handle_Store(self, stmt: ailment.Stmt.Store) -> None:
        data: MultiValues = self._expr(stmt.data)
        addr: MultiValues = self._expr(stmt.addr)
        size: int = stmt.size
        if stmt.guard is not None:
            guard = self._expr(stmt.guard)  # pylint:disable=unused-variable
        else:
            guard = None  # pylint:disable=unused-variable

        addr_v = addr.one_value()
        if addr_v is not None and not self.state.is_top(addr_v):

            if self.state.is_stack_address(addr_v):
                stack_offset = self.state.get_stack_offset(addr_v)
                if stack_offset is not None:
                    memory_location = MemoryLocation(SpOffset(self.arch.bits, stack_offset), size, endness=stmt.endness)
                else:
                    memory_location = None
            elif self.state.is_heap_address(addr_v):
                memory_location = None
            else:
                memory_location = MemoryLocation(addr_v._model_concrete.value, size, endness=stmt.endness)

            if memory_location is not None:
                self.state.kill_and_add_definition(memory_location,
                                                   self._codeloc(),
                                                   data,
                                                   endness=stmt.endness)

    def _ail_handle_Jump(self, stmt):
        _ = self._expr(stmt.target)

    def _ail_handle_ConditionalJump(self, stmt):

        cond = self._expr(stmt.condition)  # pylint:disable=unused-variable
        true_target = self._expr(stmt.true_target)  # pylint:disable=unused-variable
        false_target = self._expr(stmt.false_target)  # pylint:disable=unused-variable

        ip = Register(self.arch.ip_offset, self.arch.bytes)
        self.state.kill_definitions(ip)

        # kill all cc_ops
        if 'cc_op' in self.arch.registers:
            self.state.kill_definitions(Register(*self.arch.registers['cc_op']))
            self.state.kill_definitions(Register(*self.arch.registers['cc_dep1']))
            self.state.kill_definitions(Register(*self.arch.registers['cc_dep2']))
            self.state.kill_definitions(Register(*self.arch.registers['cc_ndep']))

    def _ail_handle_Call(self, stmt: ailment.Stmt.Call):
        self._handle_Call_base(stmt, is_expr=False)

    def _handle_Call_base(self, stmt: ailment.Stmt.Call, is_expr: bool=False):
        target = self._expr(stmt.target)  # pylint:disable=unused-variable
        codeloc = self._codeloc()

        ip = Register(self.arch.ip_offset, self.arch.bytes)
        self.state.kill_definitions(ip)

        # When stmt.args are available, used registers/stack variables are decided by stmt.args. Otherwise we fall-back
        # to using all argument registers.
        if stmt.args is not None:
            # getting used expressions from stmt.args
            used_exprs = stmt.args
        elif stmt.calling_convention is not None and (
                stmt.calling_convention.func_ty is not None or stmt.calling_convention.args is not None):
            # getting used expressions from the function prototype, its arguments, and the calling convention
            used_exprs = [ ]
            for arg_loc in stmt.calling_convention.arg_locs():
                if isinstance(arg_loc, SimRegArg):
                    used_exprs.append(Register(self.arch.registers[arg_loc.reg_name][0], arg_loc.size))
                elif isinstance(arg_loc, SimStackArg):
                    used_exprs.append(SpOffset(arg_loc.size * 8, arg_loc.stack_offset, is_base=False))
                else:
                    l.warning("_handle_Call(): Unsupported arg_loc %r.", arg_loc)
        else:
            used_exprs = None

        # All caller-saved registers will always be killed.
        if stmt.calling_convention is not None:
            cc = stmt.calling_convention
        else:
            # Fall back to the default calling convention
            l.debug("Unknown calling convention for function %s. Fall back to default calling convention.", target)
            cc = self.project.factory.cc()

        killed_vars = [ Register(*self.arch.registers[reg_name]) for reg_name in cc.CALLER_SAVED_REGS ]

        # Add uses
        if used_exprs is None:
            used_exprs = [ Register(*self.arch.registers[reg_name]) for reg_name in cc.ARG_REGS ]
        for expr in used_exprs:
            self._expr(expr)

        self.state.mark_call(codeloc, target)

        # Add definition
        return_reg_offset = None
        if not is_expr:
            if stmt.ret_expr is not None:
                if isinstance(stmt.ret_expr, ailment.Expr.Register):
                    return_reg_offset = stmt.ret_expr.reg_offset
                    return_reg_size = stmt.ret_expr.size
                    reg_atom = Register(return_reg_offset, return_reg_size)
                    top = self.state.top(return_reg_size * self.arch.byte_width)
                    self.state.kill_and_add_definition(reg_atom, codeloc, MultiValues(offset_to_values={0: {top}}))
                elif isinstance(stmt.ret_expr, ailment.Expr.Tmp):
                    tmp_atom = Tmp(stmt.ret_expr.tmp_idx, stmt.ret_expr.size)
                    top = self.state.top(stmt.ret_expr.bits)
                    self.state.kill_and_add_definition(tmp_atom, codeloc, MultiValues(offset_to_values={0: {top}}))
                else:
                    l.warning("Unsupported ret_expr type %s. Please report to GitHub.", stmt.ret_expr.__class__)

            elif cc.RETURN_VAL is not None:
                # Return value is redefined here, so it is not a dummy value
                return_reg_offset, return_reg_size = self.arch.registers[cc.RETURN_VAL.reg_name]
                self.state.kill_definitions(Register(return_reg_offset, return_reg_size))

        # Kill those ones that should be killed
        for var in killed_vars:
            if var.reg_offset == return_reg_offset:
                # Skip the return variable
                continue
            self.state.kill_definitions(var)

        # kill all cc_ops
        if 'cc_op' in self.arch.registers:
            self.state.kill_definitions(Register(*self.arch.registers['cc_op']))
            self.state.kill_definitions(Register(*self.arch.registers['cc_dep1']))
            self.state.kill_definitions(Register(*self.arch.registers['cc_dep2']))
            self.state.kill_definitions(Register(*self.arch.registers['cc_ndep']))

    def _ail_handle_Return(self, stmt: ailment.Stmt.Return):  # pylint:disable=unused-argument

        codeloc = self._codeloc()
        size = self.project.arch.bits // 8

        cc = None
        if self.state.analysis.subject.type == SubjectType.Function:
            cc = self.state.analysis.subject.content.calling_convention
            # import ipdb; ipdb.set_trace()

        if cc is None:
            # fall back to the default calling convention
            cc_cls = DEFAULT_CC.get(self.project.arch.name, None)
            if cc_cls is None:
                l.warning("Unknown default calling convention for architecture %s.", self.project.arch.name)
                cc = None
            else:
                cc = cc_cls(self.project.arch)

        if cc is not None:
            # callee-saved args
            for reg in self.arch.register_list:
                if (reg.general_purpose
                        and reg.name not in cc.CALLER_SAVED_REGS
                        and reg.name not in cc.ARG_REGS
                        and reg.vex_offset not in {self.arch.sp_offset, self.arch.bp_offset, self.arch.ip_offset, }
                        and (isinstance(cc.RETURN_VAL, SimRegArg) and reg.name != cc.RETURN_VAL.reg_name)
                ):
                    self.state.add_use(Register(reg.vex_offset, reg.size), codeloc)

        if stmt.ret_exprs:
            # Handle return expressions
            for ret_expr in stmt.ret_exprs:
                self._expr(ret_expr)
            return

        # No return expressions are available.
        # consume registers that are potentially useful

        # return value
        if cc is not None and cc.ret_val is not None:
            if isinstance(cc.ret_val, SimRegArg):
                offset = cc.ret_val._fix_offset(None, size, arch=self.project.arch)
                self.state.add_use(Register(offset, size), codeloc)
            else:
                l.error("Cannot handle CC with non-register return value location")

        # base pointer
        # TODO: Check if the stack base pointer is used as a stack base pointer in this function or not
        self.state.add_use(Register(self.project.arch.bp_offset, self.project.arch.bits // 8), codeloc)
        # We don't add sp since stack pointers are supposed to be get rid of in AIL. this is definitely a hack though
        # self.state.add_use(Register(self.project.arch.sp_offset, self.project.arch.bits // 8), codeloc)

    def _ail_handle_DirtyStatement(self, stmt: ailment.Stmt.DirtyStatement):
        # TODO: The logic below is subject to change when ailment.Stmt.DirtyStatement is changed

        if isinstance(stmt.dirty_stmt, pyvex.stmt.Dirty):
            # TODO: We need dirty helpers for a more complete understanding of clobbered registers
            tmp = stmt.dirty_stmt.tmp
            if tmp in (-1, 0xffffffff):
                return
            size = 32  # FIXME: We don't know the size.
            self.state.kill_and_add_definition(Tmp(tmp, size), self._codeloc(), None)
            self.tmps[tmp] = None
        else:
            l.warning("Unexpected type of dirty statement %s.", type(stmt.dirty_stmt))

    #
    # AIL expression handlers
    #

    def _ail_handle_BV(self, expr: claripy.ast.Base) -> MultiValues:
        return MultiValues(offset_to_values={0: {expr}})

    def _ail_handle_Tmp(self, expr: ailment.Expr.Tmp) -> MultiValues:

        self.state.add_use(Tmp(expr.tmp_idx, expr.size), self._codeloc())

        tmp = super()._ail_handle_Tmp(expr)
        if tmp is None:
            return MultiValues(offset_to_values={0: {self.state.top(expr.bits)}})
        return tmp

    def _ail_handle_CallExpr(self, expr: ailment.Stmt.Call) -> MultiValues:
        self._handle_Call_base(expr, is_expr=True)
        return MultiValues(offset_to_values={0: {self.state.top(expr.bits)}})

    def _ail_handle_Register(self, expr) -> MultiValues:

        self.state: ReachingDefinitionsState

        reg_offset = expr.reg_offset
        size = expr.size
        # bits = size * 8

        reg_atom = Register(reg_offset, size)

        # first check if it is ever defined
        try:
            value: MultiValues = self.state.register_definitions.load(reg_offset, size=size)
        except SimMemoryMissingError:
            # the value does not exist
            top = self.state.top(size * self.state.arch.byte_width)
            # annotate it
            top = self.state.annotate_with_def(top, Definition(reg_atom, ExternalCodeLocation()))
            value = MultiValues(offset_to_values={0: {top}})
            # write it back
            self.state.kill_and_add_definition(reg_atom, self._external_codeloc(), value)

        # extract Definitions
        defs: Optional[Iterable[Definition]] = None
        for vs in value.values.values():
            for v in vs:
                if defs is None:
                    defs = self.state.extract_defs(v)
                else:
                    defs = chain(defs, self.state.extract_defs(v))

        if defs is None:
            # define it right away as an external dependency
            self.state.kill_and_add_definition(reg_atom, self._external_codeloc(), value)
        else:
            codeloc = self._codeloc()
            for def_ in defs:
                self.state.add_use_by_def(def_, codeloc)

        return value

    def _ail_handle_Load(self, expr: ailment.Expr.Load) -> MultiValues:
        addrs: MultiValues = self._expr(expr.addr)

        size = expr.size
        bits = expr.bits
        if expr.guard is not None:
            guard = self._expr(expr.guard)  # pylint:disable=unused-variable
            alt = self._expr(expr.alt)  # pylint:disable=unused-variable
        else:
            guard = None  # pylint:disable=unused-variable
            alt = None  # pylint:disable=unused-variable

        # convert addrs from MultiValues to a list of valid addresses
        if len(addrs.values) == 1:
            addrs_v = next(iter(addrs.values.values()))
        else:
            top = self.state.top(bits)
            # annotate it
            dummy_atom = MemoryLocation(0, size, endness=expr.endness)
            top = self.state.annotate_with_def(top, Definition(dummy_atom, ExternalCodeLocation()))
            # add use
            self.state.add_use(dummy_atom, self._codeloc())
            return MultiValues(offset_to_values={0: {top}})

        result: Optional[MultiValues] = None
        for addr in addrs_v:
            if not isinstance(addr, claripy.ast.Base):
                continue
            if addr.concrete:
                # a concrete address
                addr = addr._model_concrete.value
                try:
                    vs: MultiValues = self.state.memory_definitions.load(addr, size=size, endness=expr.endness)
                except SimMemoryMissingError:
                    continue

                memory_location = MemoryLocation(addr, size, endness=expr.endness)
                self.state.add_use(memory_location, self._codeloc())
                result = result.merge(vs) if result is not None else vs
            elif self.state.is_stack_address(addr):
                stack_offset = self.state.get_stack_offset(addr)
                if stack_offset is not None:
                    stack_addr = self.state.live_definitions.stack_offset_to_stack_addr(stack_offset)
                    try:
                        vs: MultiValues = self.state.stack_definitions.load(stack_addr, size=size, endness=expr.endness)
                    except SimMemoryMissingError:
                        continue

                    memory_location = MemoryLocation(SpOffset(self.arch.bits, stack_offset), size, endness=expr.endness)
                    self.state.add_use(memory_location, self._codeloc())
                    result = result.merge(vs) if result is not None else vs
            else:
                l.debug('Memory address %r undefined or unsupported at pc %#x.', addr, self.ins_addr)

        if result is None:
            top = self.state.top(bits)
            # TODO: Annotate top with a definition
            result = MultiValues(offset_to_values={0: {top}})

        return result

    def _ail_handle_Convert(self, expr: ailment.Expr.Convert) -> MultiValues:
        to_conv: MultiValues = self._expr(expr.operand)
        bits = expr.to_bits
        size = bits // self.arch.byte_width

        if len(to_conv.values) == 1 and 0 in to_conv.values:
            values = to_conv.values[0]
        else:
            top = self.state.top(expr.to_bits)
            # annotate it
            dummy_atom = MemoryLocation(0, size, endness=self.arch.memory_endness)
            top = self.state.annotate_with_def(top, Definition(dummy_atom, ExternalCodeLocation()))
            # add use
            self.state.add_use(dummy_atom, self._codeloc())
            return MultiValues(offset_to_values={0: {top}})

        converted = set()
        for v in values:
            if expr.to_bits < expr.from_bits:
                conv = v[expr.to_bits - 1:0]
            elif expr.to_bits > expr.from_bits:
                conv = claripy.ZeroExt(expr.to_bits - expr.from_bits, v)
            else:
                conv = v
            converted.add(conv)

        return MultiValues(offset_to_values={0: converted})

    def _ail_handle_Reinterpret(self, expr: ailment.Expr.Reinterpret) -> MultiValues:
        _: MultiValues = self._expr(expr.operand)
        bits = expr.to_bits

        # we currently do not support floating-point operations. therefore, we return TOP directly
        reinterpreted = self.state.top(bits)

        return MultiValues(offset_to_values={0: {reinterpreted}})

    def _ail_handle_ITE(self, expr: ailment.Expr.ITE) -> MultiValues:
        _: MultiValues = self._expr(expr.cond)
        iftrue: MultiValues = self._expr(expr.iftrue)
        _: MultiValues = self._expr(expr.iffalse)
        top = self.state.top(len(iftrue))
        return MultiValues(offset_to_values={0: {top}})

    def _ail_handle_Not(self, expr: ailment.Expr.UnaryOp) -> MultiValues:
        operand: MultiValues = self._expr(expr.operand)
        bits = expr.bits

        r = None
        operand_v = operand.one_value()

        if operand_v is not None and operand_v.concrete:
            r = MultiValues(offset_to_values={0: {~operand_v}})
        else:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_BinaryOp(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        r = super()._ail_handle_BinaryOp(expr)
        if isinstance(r, ailment.Expr.BinaryOp):
            l.warning("Unimplemented operation %s.", expr.op)
            top = self.state.top(expr.bits)
            return MultiValues(offset_to_values={0: {top}})
        return r

    def _ail_handle_Add(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # adding a single value to a multivalue
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {v + expr1_v for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # adding a single value to a multivalue
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {v + expr0_v for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            # adding two single values together
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {expr0_v + expr1_v}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Sub(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # subtracting a single value from a multivalue
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {v - expr1_v for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # subtracting a single value from a multivalue
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {expr0_v - v for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {expr0_v - expr1_v}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Shr(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # each value in expr0 >> expr1_v
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {(claripy.LShR(v, expr1_v._model_concrete.value) if v.concrete else self.state.top(bits)) for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # expr0_v >> each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {(claripy.LShR(expr0_v, v._model_concrete.value) if v.concrete else self.state.top(bits)) for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {claripy.LShR(expr0_v, expr1_v._model_concrete.value)}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Sar(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # each value in expr0 >> expr1_v
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {(claripy.LShR(v, expr1_v._model_concrete.value) if v.concrete else self.state.top(bits)) for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # expr0_v >> each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {(claripy.LShR(expr0_v, v._model_concrete.value) if v.concrete else self.state.top(bits)) for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {expr0_v >> expr1_v._model_concrete.value}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Shl(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # each value in expr0 << expr1_v
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {((v << expr1_v._model_concrete.value) if v.concrete else self.state.top(bits)) for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # expr0_v >> each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {((expr0_v << v._model_concrete.value) if v.concrete else self.state.top(bits)) for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {expr0_v << expr1_v._model_concrete.value}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_And(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
            return r

        if expr0_v is None and expr1_v is not None:
            # expr1_v & each value in expr0
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {v & expr1_v for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # expr0_v & each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {expr0_v & v for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            # special handling for stack alignment
            if self.state.is_stack_address(expr0_v):
                r = MultiValues(offset_to_values={0: {expr0_v}})
            else:
                if expr0_v.concrete and expr1_v.concrete:
                    r = MultiValues(offset_to_values={0: {expr0_v & expr1_v}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Or(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # expr1_v | each value in expr0
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {v | expr1_v for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # expr0_v | each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {expr0_v | v for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {expr0_v | expr1_v}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Xor(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # expr1_v ^ each value in expr0
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {v ^ expr1_v for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # expr0_v ^ each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {expr0_v ^ v for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {expr0_v ^ expr1_v}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Concat(self, expr: ailment.Expr.BinaryOp) -> MultiValues:
        expr0: MultiValues = self._expr(expr.operands[0])
        expr1: MultiValues = self._expr(expr.operands[1])
        bits = expr.bits

        r = None
        expr0_v = expr0.one_value()
        expr1_v = expr1.one_value()

        if expr0_v is None and expr1_v is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})
        elif expr0_v is None and expr1_v is not None:
            # concatenate expr1_v with each value in expr0
            if len(expr0.values) == 1 and 0 in expr0.values:
                if all(v.concrete for v in expr0.values[0]):
                    vs = {claripy.Concat(v, expr1_v) for v in expr0.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        elif expr0_v is not None and expr1_v is None:
            # concatenate expr0_v with each value in expr1
            if len(expr1.values) == 1 and 0 in expr1.values:
                if all(v.concrete for v in expr1.values[0]):
                    vs = {claripy.Concat(expr0_v, v) for v in expr1.values[0]}
                    r = MultiValues(offset_to_values={0: vs})
        else:
            if expr0_v.concrete and expr1_v.concrete:
                r = MultiValues(offset_to_values={0: {claripy.Concat(expr0_v, expr1_v)}})

        if r is None:
            r = MultiValues(offset_to_values={0: {self.state.top(bits)}})

        return r

    def _ail_handle_Cmp(self, expr) -> MultiValues:
        op0 = self._expr(expr.operands[0])
        op1 = self._expr(expr.operands[1])

        if op0 is None: op0 = expr.operands[0]
        if op1 is None: op1 = expr.operands[1]

        top = self.state.top(expr.bits)
        return MultiValues(offset_to_values={0: {top}})

    _ail_handle_CmpF = _ail_handle_Cmp
    _ail_handle_CmpEQ = _ail_handle_Cmp
    _ail_handle_CmpNE = _ail_handle_Cmp
    _ail_handle_CmpLE = _ail_handle_Cmp
    _ail_handle_CmpLEs = _ail_handle_Cmp
    _ail_handle_CmpLT = _ail_handle_Cmp
    _ail_handle_CmpLTs = _ail_handle_Cmp
    _ail_handle_CmpGE = _ail_handle_Cmp
    _ail_handle_CmpGEs = _ail_handle_Cmp
    _ail_handle_CmpGT = _ail_handle_Cmp
    _ail_handle_CmpGTs = _ail_handle_Cmp

    def _ail_handle_Const(self, expr) -> MultiValues:
        self.state.mark_const(self._codeloc(), expr)
        if isinstance(expr.value, float):
            sort = None
            if expr.bits == 64:
                sort = FSORT_DOUBLE
            elif expr.bits == 32:
                sort = FSORT_FLOAT
            return MultiValues(offset_to_values={0: {claripy.FPV(expr.value, sort)}})
        else:
            return MultiValues(offset_to_values={0: {claripy.BVV(expr.value, expr.bits)}})

    def _ail_handle_StackBaseOffset(self, expr: ailment.Expr.StackBaseOffset) -> MultiValues:
        stack_addr = self.state.stack_address(expr.offset)
        return MultiValues(offset_to_values={0: {stack_addr}})

    def _ail_handle_DirtyExpression(self, expr: ailment.Expr.DirtyExpression) -> MultiValues:  # pylint:disable=no-self-use
        # FIXME: DirtyExpression needs .bits
        top = self.state.top(expr.bits)
        return MultiValues(offset_to_values={0: {top}})

    #
    # User defined high-level statement handlers
    #

    def _handle_function(self):
        if len(self._call_stack) + 1 > self._maximum_local_call_depth:
            l.warning('The analysis reached its maximum recursion depth.')
            return None

        defs_ip = self.state.register_definitions.get_objects_by_offset(self.arch.ip_offset)
        if len(defs_ip) != 1:
            l.error('Invalid definition(s) for IP.')
            return None

        ip_data = next(iter(defs_ip)).data
        if len(ip_data) != 1:
            l.error('Invalid number of values for IP.')
            return None

        ip_addr = ip_data.get_first_element()
        if not isinstance(ip_addr, int):
            l.error('Invalid type %s for IP.', type(ip_addr).__name__)
            return None

        is_internal = False
        ext_func_name = None
        if self.project.loader.main_object.contains_addr(ip_addr) is True:
            ext_func_name = self.project.loader.find_plt_stub_name(ip_addr)
            if ext_func_name is None:
                is_internal = True
        else:
            symbol = self.project.loader.find_symbol(ip_addr)
            if symbol is not None:
                ext_func_name = symbol.name

        if ext_func_name is not None:
            handler_name = 'handle_%s' % ext_func_name
            if hasattr(self._function_handler, handler_name):
                getattr(self._function_handler, handler_name)(self.state, self._codeloc())
            else:
                l.warning('Please implement the external function handler for %s() with your own logic.',
                          ext_func_name)
        elif is_internal is True:
            handler_name = 'handle_local_function'
            if hasattr(self._function_handler, handler_name):
                is_updated, state, visited_blocks, dep_graph = getattr(self._function_handler, handler_name)(
                    self.state,
                    ip_addr,
                    self._call_stack,
                    self._maximum_local_call_depth,
                    self._visited_blocks,
                    self._dep_graph,
                )

                if is_updated is True:
                    self.state = state
                    self._visited_blocks = visited_blocks
                    self._dep_graph = dep_graph
            else:
                l.warning('Please implement the local function handler with your own logic.')
        else:
            l.warning('Could not find function name for external function at address %#x.', ip_addr)
        return None



from typing import Optional, Set, List, Tuple, TYPE_CHECKING
import logging

import networkx

import ailment

from ..knowledge_plugins.functions import Function
from . import Analysis
from .decompiler.ailgraph_walker import AILGraphWalker
from .decompiler.ailblock_walker import AILBlockWalker

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg import CFGModel
    from angr.knowledge_plugins.xrefs import XRefManager
    from angr.analyses.decompiler.decompiler import Decompiler


_l = logging.getLogger(name=__name__)


class ProxiNodeTypes:
    String = 1
    Function = 2
    FunctionCall = 3
    Integer = 4
    Unknown = 5


class BaseProxiNode:
    """
    Base class for all nodes in a proximity graph.
    """

    def __init__(self, type_: int, ref_at: Optional[Set[int]]=None):
        self.type_ = type_
        self.ref_at = ref_at

    def __eq__(self, other):
        raise NotImplementedError()

    def __hash__(self):
        raise NotImplementedError()


class FunctionProxiNode(BaseProxiNode):

    def __init__(self, func, ref_at: Optional[Set[int]]=None):
        super().__init__(ProxiNodeTypes.Function, ref_at=ref_at)
        self.func = func

    def __eq__(self, other):
        return isinstance(other, FunctionProxiNode) and \
            other.type_ == self.type_ and \
            self.func == other.func

    def __hash__(self):
        return hash((FunctionProxiNode, self.func))


class StringProxiNode(BaseProxiNode):

    def __init__(self, addr, content, ref_at: Optional[Set[int]]=None):
        super().__init__(ProxiNodeTypes.String, ref_at=ref_at)
        self.addr = addr
        self.content = content

    def __eq__(self, other):
        return isinstance(other, StringProxiNode) and \
            other.type_ == self.type_ and \
            self.addr == other.addr

    def __hash__(self):
        return hash((StringProxiNode, self.addr))


class CallProxiNode(BaseProxiNode):

    def __init__(self, callee, ref_at: Optional[Set[int]]=None, args: Optional[Tuple[BaseProxiNode]]=None):
        super().__init__(ProxiNodeTypes.FunctionCall, ref_at=ref_at)
        self.callee = callee
        self.args = args

    def __eq__(self, other):
        return isinstance(other, CallProxiNode) and \
            other.type_ == self.type_ and \
            self.callee == other.callee and \
            self.args == other.args

    def __hash__(self):
        return hash((CallProxiNode, self.callee, self.args))


class IntegerProxiNode(BaseProxiNode):
    def __init__(self, value: int, ref_at: Optional[Set[int]]=None):
        super().__init__(ProxiNodeTypes.Integer, ref_at=ref_at)
        self.value = value

    def __eq__(self, other):
        return isinstance(other, IntegerProxiNode) and \
            self.type_ == other.type_ and \
            self.value == other.value

    def __hash__(self):
        return hash((IntegerProxiNode, self.value))


class UnknownProxiNode(BaseProxiNode):
    def __init__(self, dummy_value: str):
        super().__init__(ProxiNodeTypes.Unknown)
        self.dummy_value = dummy_value

    def __eq__(self, other):
        return isinstance(other, UnknownProxiNode) and \
            self.type_ == other.type_ and \
            self.dummy_value == other.dummy_value

    def __hash__(self):
        return hash((UnknownProxiNode, self.dummy_value))


class ProximityGraphAnalysis(Analysis):
    """
    Generate a proximity graph.
    """

    def __init__(self, func: 'Function', cfg_model: 'CFGModel', xrefs: 'XRefManager',
                 decompilation: Optional['Decompiler']=None,
                 pred_depth=1, succ_depth=1,
                 expand_funcs: Optional[Set[int]]=None):
        self._function = func
        self._cfg_model = cfg_model
        self._xrefs = xrefs
        self._decompilation = decompilation
        self._pred_depth: int = pred_depth
        self._succ_depth: int = succ_depth
        self._expand_funcs = expand_funcs.copy() if expand_funcs else None

        self.graph: Optional[networkx.DiGraph] = None

        self._work()

    def _work(self):

        self.graph = networkx.DiGraph()

        # the function
        func_proxi_node = FunctionProxiNode(self._function)

        # Process the function graph
        if not self._decompilation:
            to_expand = self._process_function(self._function, self.graph, func_proxi_node=func_proxi_node)
        else:
            to_expand = self._process_decompilation(self._function, self.graph, func_proxi_node=func_proxi_node)

        for func_node in to_expand:
            if self._expand_funcs:
                self._expand_funcs.discard(func_node.func.addr)

            subgraph = networkx.DiGraph()
            self._process_function(func_node.func, subgraph, func_proxi_node=func_node)

            # merge subgraph into the original graph
            self.graph.add_nodes_from(subgraph.nodes())
            self.graph.add_edges_from(subgraph.edges())

    def _process_strings(self, func, proxi_nodes, exclude_string_refs: Set[int]=None):
        # strings
        for v in self._cfg_model.memory_data.values():
            if exclude_string_refs and v.addr in exclude_string_refs:
                continue
            if v.sort == "string":
                xrefs = self._xrefs.xrefs_by_dst[v.addr]
                for xref in xrefs:
                    if xref.block_addr in func.block_addrs_set:
                        # include this node
                        node = StringProxiNode(v.addr, v.content, ref_at=set(x.ins_addr for x in xrefs))
                        proxi_nodes.append(node)
                        break

    def _process_function(self, func: 'Function', graph: networkx.DiGraph,
                          func_proxi_node: Optional[FunctionProxiNode]=None) -> List[FunctionProxiNode]:

        proxi_nodes: List[BaseProxiNode] = [ ]
        to_expand: List[FunctionProxiNode] = [ ]

        self._process_strings(func, proxi_nodes)

        # function calls
        for n_ in func.nodes:
            if isinstance(n_, Function):
                func_node = n_
                ref_at = set()
                for _, _, data in func.transition_graph.in_edges(func_node, data=True):
                    if 'ins_addr' in data:
                        ref_at.add(data['ins_addr'])
                if self._expand_funcs and func_node.addr in self._expand_funcs:  # pylint:disable=unsupported-membership-test
                    node = FunctionProxiNode(func_node, ref_at=ref_at)
                    to_expand.append(node)
                else:
                    node = CallProxiNode(func_node, ref_at=ref_at)
                proxi_nodes.append(node)

        # add it to the graph
        graph.add_node(func_proxi_node)
        for pn in proxi_nodes:
            graph.add_edge(func_proxi_node, pn)

        return to_expand

    def _process_decompilation(self, func: 'Function', graph: networkx.DiGraph,
                               func_proxi_node: Optional[FunctionProxiNode]=None) -> List[FunctionProxiNode]:
        proxi_nodes: List[BaseProxiNode] = [ ]
        to_expand: List[FunctionProxiNode] = [ ]

        # dedup
        string_refs: Set[int] = set()
        func_calls: Set[CallProxiNode] = set()

        # Walk the clinic structure to dump string references and function calls
        ail_graph = self._decompilation.clinic.graph

        def _handle_Call(stmt_idx: int, stmt: ailment.Stmt.Call, block: Optional[ailment.Block]):  # pylint:disable=unused-argument
            func_node = self.kb.functions[stmt.target.value]
            ref_at = { stmt.ins_addr }

            # extract arguments
            args = [ ]
            if stmt.args:
                for arg in stmt.args:
                    if isinstance(arg, ailment.Expr.Const):
                        # is it a reference to a string?
                        if arg.value in self._cfg_model.memory_data:
                            md = self._cfg_model.memory_data[arg.value]
                            if md.sort == "string":
                                # Yes!
                                args.append(StringProxiNode(arg.value, md.content))
                                string_refs.add(arg.value)
                        else:
                            # not a string. present it as a constant integer
                            args.append(IntegerProxiNode(arg.value, None))
                    else:
                        args.append(UnknownProxiNode("_"))

            if self._expand_funcs and func_node.addr in self._expand_funcs:  # pylint:disable=unsupported-membership-test
                node = FunctionProxiNode(func_node, ref_at=ref_at)
                to_expand.append(node)
            else:
                node = CallProxiNode(func_node, ref_at=ref_at, args=tuple(args) if args is not None else None)

            if node in func_calls:
                return
            func_calls.add(node)
            proxi_nodes.append(node)

        def _handle_CallExpr(self, expr_idx: int, expr: ailment.Stmt.Call, stmt_idx: int, stmt: ailment.Stmt.Statement, block: Optional[ailment.Block]):  # pylint:disable=unused-argument
            func_node = self.kb.functions[expr.target.value]
            ref_at = { stmt.ins_addr }
            if self._expand_funcs and func_node.addr in self._expand_funcs:
                node = FunctionProxiNode(func_node, ref_at=ref_at)
                to_expand.append(node)
            else:
                node = CallProxiNode(func_node, ref_at=ref_at)
            proxi_nodes.append(node)

        stmt_handlers = {
            ailment.Stmt.Call: _handle_Call,
        }
        expr_handlers = {
            ailment.Stmt.Call: _handle_CallExpr,
        }

        block_walker = AILBlockWalker(stmt_handlers=stmt_handlers, expr_handlers=expr_handlers)

        def _graph_walker_handler(node):
            block_walker.walk(node)

        AILGraphWalker(ail_graph, _graph_walker_handler, replace_nodes=False).walk()


        # add strings references that are not used in function calls
        self._process_strings(func, proxi_nodes, exclude_string_refs=string_refs)

        # add it to the graph
        graph.add_node(func_proxi_node)
        for pn in proxi_nodes:
            graph.add_edge(func_proxi_node, pn)

        return to_expand


from angr.analyses import AnalysesHub
AnalysesHub.register_default('Proximity', ProximityGraphAnalysis)




import pyvex

from .errors import SimSlicerError


class SimLightState(object):
    def __init__(self, temps=None, regs=None, stack_offsets=None, options=None):
        self.temps = temps if temps is not None else set()
        self.regs = regs if regs is not None else set()
        self.stack_offsets = stack_offsets if stack_offsets is not None else set()
        self.options = {} if options is None else options


class SimSlicer(object):
    """
    A super lightweight intra-IRSB slicing class.
    """
    def __init__(self, arch, statements, target_tmps=None, target_regs=None, target_stack_offsets=None,
                 inslice_callback=None, inslice_callback_infodict=None):
        self._arch = arch
        self._statements = statements
        self._target_tmps = target_tmps if target_tmps is not None else set()
        self._target_regs = target_regs if target_regs is not None else set()
        self._target_stack_offsets = target_stack_offsets if target_stack_offsets is not None else set()

        self._inslice_callback = inslice_callback

        # It could be accessed publicly
        self.inslice_callback_infodict = inslice_callback_infodict

        self.stmts = [ ]
        self.stmt_indices = [ ]
        self.final_regs = set()
        self.final_stack_offsets = set()

        if not self._target_tmps and not self._target_regs and not self._target_stack_offsets:
            raise SimSlicerError('You must specify at least one of the following: "'
                                 'target temps, target registers, and/or target stack offsets.')

        # convert target registers to base registers
        target_base_regs = set()
        for target_reg in self._target_regs:
            base_reg = self._arch.get_base_register(target_reg)
            if base_reg is None:
                target_base_regs.add(target_reg)
            else:
                target_base_regs.add(base_reg[0])
        self._target_regs = target_base_regs

        self._aliases = { }

        self._alias_analysis()
        self._slice()

    def _alias_analysis(self, mock_sp=True, mock_bp=True):
        """
        Perform a forward execution and perform alias analysis. Note that this analysis is fast, light-weight, and by no
        means complete. For instance, most arithmetic operations are not supported.

        - Depending on user settings, stack pointer and stack base pointer will be mocked and propagated to individual
          tmps.

        :param bool mock_sp: propagate stack pointer or not
        :param bool mock_bp: propagate stack base pointer or not
        :return: None
        """

        state = SimLightState(
            regs={
                self._arch.sp_offset: self._arch.initial_sp,
                self._arch.bp_offset: self._arch.initial_sp + 0x2000, # TODO: take care of the relation between sp and bp
            },
            temps={},
            options={
            'mock_sp': mock_sp,
            'mock_bp': mock_bp,
            }
        )

        for stmt_idx, stmt in list(enumerate(self._statements)):
            self._forward_handler_stmt(stmt, state)

    #
    # Forward execution IRStmt handlers
    #

    def _forward_handler_stmt(self, stmt, state):
        """

        :param stmt:
        :param SimLightState state:
        :return:
        """

        funcname = "_forward_handler_stmt_%s" % type(stmt).__name__

        if hasattr(self, funcname):
            getattr(self, funcname)(stmt, state)

    def _forward_handler_stmt_WrTmp(self, stmt, state):
        tmp = stmt.tmp

        val = self._forward_handler_expr(stmt.data, state)

        if val is not None:
            state.temps[tmp] = val
            self._aliases[tmp] = val

    #
    # Forward execution IRExpr handlers
    #

    def _forward_handler_expr(self, expr, state):
        """

        :param stmt:
        :param SimLightState state:
        :return:
        """

        funcname = "_forward_handler_expr_%s" % type(expr).__name__

        if hasattr(self, funcname):
            return getattr(self, funcname)(expr, state)

        return None

    def _forward_handler_expr_Get(self, expr, state):
        reg = expr.offset

        if state.options['mock_sp'] and reg == self._arch.sp_offset:
            return state.regs[reg]

        elif state.options['mock_bp'] and reg == self._arch.bp_offset:
            return state.regs[reg]

        return None

    def _forward_handler_expr_RdTmp(self, expr, state):
        tmp = expr.tmp

        if tmp in state.temps:
            return state.temps[tmp]

        return None

    def _forward_handler_expr_Const(self, expr, state):

        return expr.con.value

    def _forward_handler_expr_Binop(self, expr, state):

        funcname = "_forward_handler_expr_binop_%s" % expr.op.strip("Iop_")

        if hasattr(self, funcname):
            op0_val = self._forward_handler_expr(expr.args[0], state)
            op1_val = self._forward_handler_expr(expr.args[1], state)
            if op0_val is not None and op1_val is not None:
                return getattr(self, funcname)(op0_val, op1_val, state)

        return None

    def _forward_handler_expr_binop_Add64(self, op0, op1, state):

        return (op0 + op1) & (2 ** 64 - 1)

    def _forward_handler_expr_binop_Add32(self, op0, op1, state):

        return (op0 + op1) & (2 ** 32 - 1)

    #
    # Backward slicing
    #

    def _slice(self):
        """
        Slice it!
        """

        regs = set(self._target_regs)
        tmps = set(self._target_tmps)
        stack_offsets = set(self._target_stack_offsets)

        state = SimLightState(regs=regs, temps=tmps, stack_offsets=stack_offsets)

        for stmt_idx, stmt in reversed(list(enumerate(self._statements))):
            if self._backward_handler_stmt(stmt, state):
                self.stmts.insert(0, stmt)
                self.stmt_indices.insert(0, stmt_idx)

                if self._inslice_callback:
                    self._inslice_callback(stmt_idx, stmt, self.inslice_callback_infodict)

            if not regs and not tmps and not stack_offsets:
                break

        self.final_regs = state.regs
        self.final_stack_offsets = state.stack_offsets

    #
    # Backward slice IRStmt handlers
    #

    def _backward_handler_stmt(self, stmt, state):
        funcname = "_backward_handler_stmt_%s" % type(stmt).__name__

        in_slice = False
        if hasattr(self, funcname):
            in_slice = getattr(self, funcname)(stmt, state)

        return in_slice

    def _backward_handler_stmt_WrTmp(self, stmt, state):
        tmp = stmt.tmp

        if tmp not in state.temps:
            return False

        state.temps.remove(tmp)

        self._backward_handler_expr(stmt.data, state)

        return True

    def _backward_handler_stmt_Put(self, stmt : pyvex.IRStmt.Put, state):
        reg = stmt.offset
        # convert it to its base register
        base_reg = self._arch.get_base_register(reg)
        if base_reg is not None:
            reg = base_reg[0]

        if reg in state.regs:
            state.regs.remove(reg)

            self._backward_handler_expr(stmt.data, state)

            return True

        else:
            return False

    def _backward_handler_stmt_Store(self, stmt, state):

        addr = stmt.addr

        if type(addr) is pyvex.IRExpr.RdTmp:
            tmp = addr.tmp

            if tmp in self._aliases:
                # We know its value
                concrete_addr = self._aliases[tmp]
                if concrete_addr in state.stack_offsets:
                    # It's written at this statement
                    state.stack_offsets.remove(concrete_addr)
                    self._backward_handler_expr(addr, state)
                    self._backward_handler_expr(stmt.data, state)

                    return True

        return False

    def _backward_handler_stmt_LoadG(self, expr, state):

        if expr.dst not in state.temps:
            return False

        state.temps.remove(expr.dst)

        self._backward_handler_expr(expr.guard, state)
        self._backward_handler_expr(expr.addr, state)
        self._backward_handler_expr(expr.alt, state)

        return True

    #
    # Backward slice IRExpr handlers
    #

    def _backward_handler_expr(self, expr, state):
        funcname = "_backward_handler_expr_%s" % type(expr).__name__
        if hasattr(self, funcname):
            getattr(self, funcname)(expr, state)

    def _backward_handler_expr_RdTmp(self, expr, state):
        tmp = expr.tmp

        state.temps.add(tmp)

    def _backward_handler_expr_Get(self, expr, state):
        reg = expr.offset
        # convert it to its base register
        base_reg = self._arch.get_base_register(reg)
        if base_reg is not None:
            reg = base_reg[0]

        state.regs.add(reg)

    def _backward_handler_expr_Load(self, expr, state):
        addr = expr.addr

        if type(addr) is pyvex.IRExpr.RdTmp:
            self._backward_handler_expr(addr, state)

            # Do we know the concrete value of this tmp?
            tmp = addr.tmp
            if tmp in self._aliases:
                # awesome!
                state.stack_offsets.add(self._aliases[tmp])

    def _backward_handler_expr_Unop(self, expr, state):
        arg = expr.args[0]

        if type(arg) is pyvex.IRExpr.RdTmp:
            self._backward_handler_expr(arg, state)

    def _backward_handler_expr_CCall(self, expr, state):

        for arg in expr.args:
            if type(arg) is pyvex.IRExpr.RdTmp:
                self._backward_handler_expr(arg, state)

    def _backward_handler_expr_Binop(self, expr, state):

        for arg in expr.args:
            if type(arg) is pyvex.IRExpr.RdTmp:
                self._backward_handler_expr(arg, state)

    def _backward_handler_expr_ITE(self, expr, state):

        self._backward_handler_expr(expr.cond, state)
        self._backward_handler_expr(expr.iftrue, state)
        self._backward_handler_expr(expr.iffalse, state)
